/*
 * 20200110-07-9999.c
 *	浮動小数点数のキュー(queue)
 */

#include <stdio.h>
#include <malloc.h>		/* malloc を利用する場合 */

/*
 *  浮動小数点数のキュー(queue)
 *
 * 利用方法
 *		コンパイル
 *			cc -o BASENAME.exe FILENAME
 *		実行
 *			./BASENAME.exe
 */

/*
 * Queue は、List の最後に追加し、先頭の要素を削除する事で実現
 *
 *		浮動小数点数 queue : [ 1.0, 2.0, .., 10.0 ] の表現
 *
 *								DCell
 *								+-------+	+-------+			+-------+
 *					next	+->	|	*----->	|	*-----> ... -->	| NULL	|
 *					value	|	|		|	|		|		+->	|		|
 *							|	+-------+	+-------+	...	|	+-------+
 *							|	|	1.0	|	|	2.0	|		|	|	10.0|
 *							|	+-------+	+-------+		|	+-------+
 *			DQueue			|								|
 *			+-----------+	|								|
 *	head	|		*-------+								|
 *			+-----------+									|
 *	tail	|		*---------------------------------------+
 *			+-----------+
 *
 *		空な浮動小数点数 queue : [] の表現
 *			空な浮動小数点数 queue は、head/tail 共に NULL とする
 *
 *			DQueue
 *			+-----------+
 *	head	|	NULL	|
 *			+-----------+
 *	tail	|	NULL	|
 *			+-----------+
 *
 */

typedef struct icell {
	struct icell *next;	/* 次のセルへのポインター */
	double	value;		/* このセルが持つ値(浮動小数点数) */
} DCell;

typedef struct {
	DCell *head;		/* リストの先頭の要素へのポインター */
	DCell *tail;		/* リストの最後の要素へのポインター */
} DQueue;

/*
*/

#define	NORMAL	(0)		/* 処理に成功した場合の値 */
#define	ERROR	(-1)	/* 処理に失敗した場合の値 */

/*
*/

#define	FALSE	(0)			/* 論理値 (偽値) */
#define	TRUE	(!FALSE)	/* 論理値 (真値) */

/*
 * alloc_dcell
 *	指定した浮動小数点数値(data) を保持する dcell を作成する
 *	return 値
 *		作成された dcell へのポインター値
 *			alloc に失敗した場合は NULL 値が返る
 */

DCell *alloc_dcell ( int data ) {
	DCell *result = (DCell *)malloc(sizeof(DCell));

	if ( result != NULL ) {
		/* 次の要素 (next) は NULL にする */
		result -> next = NULL;

		/* 保持する値 (value) は 指定された値 (data) にする */

	/*
	**	 この部分を完成させなさい
	*/

	}

	return result;
}

/*
 * free_dcell
 *	指定した浮動小数点数値を保持する dcell (icp) を開放する
 *	return 値
 *		開放に成功したかどうか ( 成功 -> NORMAL / 失敗 -> ERROR )
 */

int free_dcell ( DCell *dcp ) {
	int result = ERROR;

	if ( dcp != NULL ) {
		free ( dcp );
		result = NORMAL;
	}

	return result;
}

/*
 * alloc_dqueue
 *	空っぽな浮動小数点数 queeue (dqueue) を作成する
 *	return 値
 *		作成された dqueue へのポインター値
 *			alloc に失敗した場合は NULL 値が返る
 */

DQueue *alloc_dqueue ( void ) {
	DQueue *result = (DQueue *)malloc(sizeof(DQueue));

	if ( result != NULL ) {
		/* 空な浮動小数点数 queue は、先頭も最後も NULL にする */

	/*
	**	 この部分を完成させなさい
	*/

	}

	return result;
}

/*
 * is_empty
 *	指定した浮動小数点数 queue が空かどうかを判定する
 *	return 値
 *		空の場合		TRUE
 *		空でない場合	TRUE 以外
 */

int is_empty ( DQueue *dqp ) {

	if ( dqp != NULL ) {
		if ( dqp -> head == NULL ) {
		   return TRUE;
		}
	}

	return FALSE;
}

/*
 * enque
 *	指定された 浮動小数点数 queue (dqp) に指定された浮動小数点数値 (data) を追加する
 *	return 値
 *		追加できた場合	NORMAL
 *		それ以外		ERROR
 */

int enque ( DQueue *dqp, double data ) {
	int result = ERROR;	/* 途中で、何か上手く行かなければ ERROR */

	if ( dqp != NULL ) {
		DCell *dcp = alloc_dcell ( data );	/* 新しい DCell を作成 */

		if ( dcp != NULL ) {				/* 作成できたら */
			if ( is_empty ( dqp ) ) {		/* 空っぽの場合 */
				dqp -> head = dcp;			/* 先頭も新しい要素 */
			} else {
				dqp -> tail -> next = dcp;	/* 最後の要素の次に新しい要素を追加 */
			}
							/* 最後の要素は、追加した要素 */

	/*
	**	 この部分を完成させなさい
	*/

			result = NORMAL;				/* 追加に成功 (全て OK) */
		}
	}

	return result;
}

/*
 * dequeue
 *	指定された 浮動小数点数 queue (dqp) から、要素を取り出す
 *	return 値
 *		取り出せた場合		取り出した値
 *		取り出せない場合	ERROR
 *		<<注意>>
 *			浮動小数点数 queue の中に ERROR があると、
 *			取り出しに成功したか失敗したか区別できない (不適切な設計 !!)
 */

int deque ( DQueue *dqp ) {
	int result = ERROR;

	if ( dqp != NULL ) {
		if ( !is_empty ( dqp ) ) {		/* queue が空でない */
			DCell *top = dqp -> head;	/* 先頭の要素(DCell)を取り出す */

			result = top -> value;		/* 返す値は、先頭にある */

			if ( ( dqp -> head = top -> next) == NULL ) {
										/* 新しい先頭は、先頭の次の要素 */
				/* queue が空になった場合は tail も NULL にする */

	/*
	**	 この部分を完成させなさい
	*/

			}
			free ( top );				/* 先頭の要素は開放 */
		}
	}

	return result;
}

/*
 * free_que
 *	指定された 浮動小数点数 queue (dqp) を全て開放する
 */

void free_que ( DQueue *dqp ) {

	if ( dqp != NULL ) {
		DCell *dcp = dqp -> head;	/* 先頭の要素を、次の処理対象に */
		DCell *next;				/* 次の要素へのポインタ値を保持 */

		while ( dcp != NULL ) {		/* 未だ、要素があれば.. */
			next = dcp -> next;		/* 次のために、次の要素へのポインタ値をバックアップ */
			free_dcell ( dcp );		/* 注目している要素を開放 */
			/* 次の要素を、次の処理対象にする */

	/*
	**	 この部分を完成させなさい
	*/

		}

		free ( dqp );				/* queue そのものを開放 */
	}
}

/*
 * print_que
 *	指定された浮動小数点数 que の内容を表示する
 */

void print_que ( DQueue *dqp ) {
	DCell *dcp = dqp -> head;

	printf ( "[" );
	for ( dcp = dqp -> head; dcp != NULL; dcp = dcp -> next ) {
		/* 現在着目している要素 (dcp) が保持する値を表示 */

	/*
	**	 この部分を完成させなさい
	*/

		if ( dcp -> next != NULL ) {		/* 次の要素があれば .. */
			printf ( "," );				/* 「,」を出力 */
		} else {
			printf ( " " );				/* 「 」を出力 */
		}
	}
	printf ( "]" );
}

/*
 * main
 */

int main ( int argc, char *argv[] ) {
	DQueue *dqp = alloc_dqueue();			/* 空の queue の作成 */

	print_que ( dqp ); putchar ( '\n' );	/* 空の queue [] */

	enque ( dqp, /* q: ここ */ );
	print_que ( dqp ); putchar ( '\n' );

	enque ( dqp, 1.0 );						/* que に 1.0 を追加 */
	enque ( dqp, /* p: ここ */ );
	print_que ( dqp ); putchar ( '\n' );

	printf ( "%d\n", deque ( dqp ) );		/* 先頭の要素を削除 */
	print_que ( dqp ); putchar ( '\n' );

	free_que ( dqp );						/* que の開放 */

	return 0;
}
